/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

#ifdef HAS_IRODS

#include "irods_reli.h"

#include "rodsClient.h"
#include "rodsPath.h"
#include "miscUtil.h"

#include "hash_table.h"
#include "debug.h"
#include "stringtools.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

static struct hash_table *connect_cache = 0;
static rodsEnv irods_env;
static int got_irods_env = 0;

struct irods_file {
	int fd;
	INT64_T offset;
	char *server;
};

struct irods_server {
	rcComm_t *conn;
	time_t lastused;
};

#define DISCONNECT_TIMEOUT 300

static rcComm_t * irods_reli_connect_cache( const char *hostport )
{
	struct irods_server *server;
	rcComm_t *conn;
	rErrMsg_t errmsg;
	char host[PFS_PATH_MAX];
	int port;

	if(!hostport[0]) {
		errno = EACCES;
		return 0;
	}

	if(!connect_cache) connect_cache = hash_table_create(0,0);

	time_t current = time(0);

	server = (struct irods_server*) hash_table_lookup(connect_cache,hostport);
	if(server) {
		if((current-server->lastused)>DISCONNECT_TIMEOUT) {
			debug(D_IRODS,"discarding stale connection to %s",hostport);
			rcDisconnect(server->conn);
			free(server);
			server = 0;
		} else {
			return server->conn;
		}
	}

	if(!got_irods_env) {
		rodsLogLevel(100);
		if(getRodsEnv(&irods_env)<0) {
			debug(D_IRODS,"couldn't load irods environment!");
			return 0;
		}
		got_irods_env = 1;
	}

	debug(D_IRODS,"connecting to %s",hostport);

	sscanf(hostport,"%[^:]:%d",host,&port);

	conn = rcConnect(host,port,irods_env.rodsUserName,irods_env.rodsZone,1,&errmsg);
	if(!conn) {
		if(errno==EINPROGRESS) {
			errno = ECONNREFUSED;
		}
		debug(D_IRODS,"couldn't connect: %s\n",strerror(errno));
		return 0;
	}

	debug(D_IRODS,"logging in...");

	if(clientLogin(conn)!=0) {
		debug(D_IRODS,"unable to login as %s",irods_env.rodsUserName);
		rcDisconnect(conn);
		return 0;
	}

	server = malloc(sizeof(*server));
	server->conn = conn;
	server->lastused = current;

	hash_table_insert(connect_cache,hostport,server);

	return conn;
}

struct irods_file * irods_reli_open ( const char *server, const char *path, int flags, int mode )
{
	struct irods_file *file;
	dataObjInp_t request;
	int result;

	rcComm_t *conn = irods_reli_connect_cache(server);
	if(!conn) return 0;

	memset(&request,0,sizeof(request));
	strcpy(request.objPath,path);
	request.openFlags = flags;

	debug(D_IRODS,"rcDataObjOpen %s %s",server,path);
	result = rcDataObjOpen (conn,&request);
	debug(D_IRODS,"= %d",result);

	if(result<0 && flags&O_CREAT) {
		memset(&request,0,sizeof(request));
		strcpy(request.objPath,path);
		request.createMode = mode;
		request.openFlags = flags;
		request.dataSize = -1;

		addKeyVal (&request.condInput, DATA_TYPE_KW, "generic");

		if(irods_env.rodsDefResource[0]) {
			addKeyVal (&request.condInput, RESC_NAME_KW,irods_env.rodsDefResource);
		}

		debug(D_IRODS,"rcDataObjCreate %s %s",server,path);
		result = rcDataObjCreate (conn,&request);
		debug(D_IRODS,"= %d",result);
	}

	if(result<0) {
		errno = getUnixErrno(result);
		return 0;
	}

	if(flags&O_TRUNC) {
		irods_reli_truncate(server,path,0);
	}

	file = malloc(sizeof(*file));
	file->fd = result;
	file->offset = 0;
	file->server = strdup(server);

	return file;
}

static int irods_reli_lseek_if_needed( struct irods_file *file, INT64_T offset )
{
	fileLseekInp_t request;
	fileLseekOut_t *response=0;
	int result;

	if(file->offset==offset)  return 0;

	rcComm_t *conn = irods_reli_connect_cache(file->server);
	if(!conn) return -1;

	request.fileInx = file->fd;
	request.offset  = offset;
	request.whence  = SEEK_SET;

	debug(D_IRODS,"rcDataObjLseek %s %d %lld",file->server,file->fd,offset);
	result = rcDataObjLseek(conn,&request,&response);
	debug(D_IRODS,"= %d",result);

	if(result<0) {
		errno = getUnixErrno(result);
		return -1;
	}

	free(response);
	file->offset = offset;
	return 0;
}

int irods_reli_pread ( struct irods_file *file, char *data, int length, INT64_T offset )
{
	int result;
	dataObjReadInp_t request;
	bytesBuf_t       response;

	rcComm_t *conn = irods_reli_connect_cache(file->server);
	if(!conn) return -1;

	if(irods_reli_lseek_if_needed(file,offset)<0) return -1;

	request.l1descInx = file->fd;
	request.len = length;

	response.buf = data;
	response.len = length;

	debug(D_IRODS,"rcDataObjRead %s %d %d",file->server,file->fd,length);
	result = rcDataObjRead(conn,&request,&response);
	debug(D_IRODS,"= %d",result);

	if(result<0) {
		errno = getUnixErrno(result);
		return -1;
	}

	file->offset += result;

	return result;
}

int irods_reli_pwrite    ( struct irods_file *file, const char *data, int length, INT64_T offset )
{
	int result;
	dataObjWriteInp_t request;
	bytesBuf_t        response;

	rcComm_t *conn = irods_reli_connect_cache(file->server);
	if(!conn) return -1;

	if(irods_reli_lseek_if_needed(file,offset)<0) return -1;

	request.l1descInx = file->fd;
	request.len = length;

	response.buf = (char*)data;
	response.len = length;

	debug(D_IRODS,"rcDataObjWrite %s %d %d",file->server,file->fd,length);
	result = rcDataObjWrite(conn,&request,&response);
	debug(D_IRODS,"= %d",result);

	if(result<0) {
		errno = getUnixErrno(result);
		return -1;
	}

	file->offset += result;

	return result;
}

int irods_reli_close    ( struct irods_file *file )
{
	int result;
	dataObjCloseInp_t request;

	rcComm_t *conn = irods_reli_connect_cache(file->server);
	if(!conn) return -1;

	memset(&request,0,sizeof(request));
	request.l1descInx = file->fd;

	debug(D_IRODS,"rcDataObjClose %s %d",file->server,file->fd);
	result = rcDataObjClose(conn,&request);
	debug(D_IRODS,"= %d",result);

	free(file->server);
	free(file);

	return 0;
}


int irods_reli_stat ( const char *server, const char *path, struct pfs_stat *info )
{
	int result;
	dataObjInp_t request;
	rodsObjStat_t *response = 0;

	/* Special case: /irods looks like an empty dir */

	if(!server[0]) {
		memset(info,0,sizeof(*info));
		info->st_ino = 2;
		info->st_dev = -1;
		info->st_mode = S_IFDIR | 0755;
		info->st_size = 4096;
		info->st_blocks = 8;
		info->st_nlink = 1;
		return 0;
	}

	rcComm_t *conn = irods_reli_connect_cache(server);
	if(!conn) return -1;

	memset(&request,0,sizeof(request));
	strcpy(request.objPath,path);

	debug(D_IRODS,"rcObjStat %s %s",server,path);
	result = rcObjStat (conn,&request,&response);
	debug(D_IRODS,"= %d",result);

	if(result<0) {
		errno = getUnixErrno(result);
		return -1;
	}

	memset(info,0,sizeof(*info));

	info->st_ino = hash_string(path);
	info->st_dev = -1;

	if (response->objType == COLL_OBJ_T) {
		info->st_mode = S_IFDIR | 0755;
		info->st_size = 4096;
		info->st_blocks = 8;
		info->st_ctime = atoi (response->createTime);
		info->st_mtime = atoi (response->modifyTime);
		info->st_nlink = 1;
	} else {
		info->st_mode = S_IFREG | 0755;
		info->st_size = response->objSize;
		info->st_blocks = info->st_size/512+1;
		info->st_nlink = response->numCopies;
	}

	info->st_blksize = 4096;
	info->st_ctime = atoi(response->createTime);
	info->st_mtime = atoi(response->modifyTime);
	info->st_atime = atoi(response->modifyTime);

	free(response);

	return 0;
}

int irods_reli_getdir( const char *server, const char *path, void (*callback) ( const char *path, void *arg ), void *arg )
{
	int i;
	int result;
	int keepgoing;

	genQueryInp_t query_in;
	genQueryOut_t *query_out = 0;
	rodsArguments_t args;
	sqlResult_t *value;

	/* Special case: /irods looks like an empty dir */

	if(!server[0]) {
		callback(".",arg);
		callback("..",arg);
		return 0;
	}

	rcComm_t *conn = irods_reli_connect_cache(server);
	if(!conn) return -1;

	/* First we set up a query for file names */

	memset(&query_in,0,sizeof(query_in));
	memset(&args,0,sizeof(args));
	query_out = 0;

	debug(D_IRODS,"queryDataObjInColl %s %s",server,path);
	result = queryDataObjInColl(conn,(char*)path,&args,&query_in,&query_out);
	debug(D_IRODS,"= %d",result);

	if(result<0 && result!=CAT_NO_ROWS_FOUND) {
		errno = ENOTDIR;
		return -1;
	}

	callback(".",arg);
	callback("..",arg);

	while(result>=0) {
		value = getSqlResultByInx(query_out,COL_DATA_NAME);
		if(!value) break;

		char *subname = value->value;
			
		for(i=0;i<query_out->rowCnt;i++) {
			callback(subname,arg);
			subname += value->len;
		}

		keepgoing = query_out->continueInx;
		freeGenQueryOut(&query_out);

		if(keepgoing>0) {
			query_in.continueInx = keepgoing;
			debug(D_IRODS,"rcGenQuery %s",server);
			result = rcGenQuery(conn,&query_in,&query_out);
			debug(D_IRODS,"= %d",result);
		} else {
			break;
		}
	}

	/* Now we set up a SECOND query for directory names */

	memset(&query_in,0,sizeof(query_in));
	memset(&args,0,sizeof(args));
	query_out = 0;

	debug(D_IRODS,"queryCollInColl %s %s",server,path);
	result = queryCollInColl(conn,(char*)path,&args,&query_in,&query_out);
	debug(D_IRODS,"= %d",result);

	if(result<0 && result!=CAT_NO_ROWS_FOUND) {
		return 0;
	}

	while(result>=0) {
		value = getSqlResultByInx(query_out,COL_COLL_NAME);
		if(!value) break;
	
		char *subname = value->value;
			
		for(i=0;i<query_out->rowCnt;i++) {
			callback(subname,arg);
			subname += value->len;
		}

		keepgoing = query_out->continueInx;
		freeGenQueryOut(&query_out);

		if(keepgoing>0) {
			query_in.continueInx = keepgoing;
			debug(D_IRODS,"rcGenQuery %s",server);
			result = rcGenQuery(conn,&query_in,&query_out);
			debug(D_IRODS,"= %d",result);
		} else {
			break;
		}
	}

	return 0;
}

int irods_reli_statfs   ( const char *server, const char *path, struct pfs_statfs *info )
{
	memset(info,0,sizeof(info));

	info->f_blocks = 10000000;
	info->f_bavail = 10000000;
	info->f_bsize  = 4096;
	info->f_bfree  = 10000000;
	info->f_files  = 10000000;
	info->f_ffree  = 10000000;

	return 0;
}

int irods_reli_mkdir ( const char *server, const char *path )
{
	collInp_t request;
	int result;

	rcComm_t *conn = irods_reli_connect_cache(server);
	if(!conn) return -1;

	memset(&request,0,sizeof(request));
	strcpy(request.collName,path);

	debug(D_IRODS,"rcCollCreate %s %s",server,path);
	result = rcCollCreate(conn,&request);
	debug(D_IRODS,"= %d",result);

	if(result<0) {
		errno = getUnixErrno(result);
		return -1;
	}

	return result;
}

int irods_reli_unlink   ( const char *server, const char *path )
{
	dataObjInp_t request;
	int result;

	rcComm_t *conn = irods_reli_connect_cache(server);
	if(!conn) return -1;

	memset(&request,0,sizeof(request));
	strcpy(request.objPath,path);

	addKeyVal (&request.condInput, FORCE_FLAG_KW, "");

	debug(D_IRODS,"rcUnlink %s %s",server,path);
	result = rcDataObjUnlink(conn,&request);
	debug(D_IRODS,"= %d",result);

	clearKeyVal (&request.condInput);

	if(result<0) {
		errno = getUnixErrno(result);
		return -1;
	}

	return result;
}

int irods_reli_rmdir ( const char *server, const char *path )
{
	collInp_t request;
	int result;

	rcComm_t *conn = irods_reli_connect_cache(server);
	if(!conn) return -1;

	memset(&request,0,sizeof(request));
	strcpy(request.collName,path);

	addKeyVal (&request.condInput, FORCE_FLAG_KW, "");

	debug(D_IRODS,"rcRmColl %s %s",server,path);
	result = rcRmColl(conn,&request);
	debug(D_IRODS,"= %d",result);

	clearKeyVal (&request.condInput);

	if(result<0) {
		errno = getUnixErrno(result);
		return -1;
	}

	return result;
}

int irods_reli_rename   ( const char *server, const char *path, const char *newpath )
{
	dataObjCopyInp_t request;
	int result;

	/* rcDataObjRename does not have the Unix semantics of atomically
	   destroying the target file.  So, we must do it ourselves.
	*/

	irods_reli_unlink(server,newpath);

	rcComm_t *conn = irods_reli_connect_cache(server);
	if(!conn) return -1;

	memset(&request,0,sizeof(request));
	strcpy(request.srcDataObjInp.objPath,path);
	strcpy(request.destDataObjInp.objPath,newpath);
	request.srcDataObjInp.oprType  = RENAME_DATA_OBJ;
	request.destDataObjInp.oprType = RENAME_DATA_OBJ;

	debug(D_IRODS,"rcDataObjRename %s %s %s",server,path,newpath);
	result = rcDataObjRename(conn,&request);
	debug(D_IRODS,"= %d",result);	

	if(result<0) {
		errno = getUnixErrno(result);
		return -1;
	}

	return result;
}

int irods_reli_truncate ( const char *server, const char *path, INT64_T length )
{
	dataObjInp_t request;
	int result;

	rcComm_t *conn = irods_reli_connect_cache(server);
	if(!conn) return -1;

	memset(&request,0,sizeof(request));
	strcpy(request.objPath,path);
	request.dataSize = length;

	debug(D_IRODS,"rcDataObjTruncate %s %s",server,path);
	result = rcDataObjTruncate(conn,&request);
	debug(D_IRODS,"= %d",result);

	if(result<0) {
		errno = getUnixErrno(result);
		return -1;
	}

	return 0;
}

static unsigned char hex_to_nybble( char h )
{
	switch(h) {
		case '0':	return 0;
		case '1':	return 1;
		case '2':	return 2;
		case '3':	return 3;
		case '4':	return 4;
		case '5':	return 5;
		case '6':	return 6;
		case '7':	return 7;	
		case '8':	return 8;
		case '9':	return 9;
		case 'a':	return 10;
		case 'b':	return 11;
		case 'c':	return 12;
		case 'd':	return 13;
		case 'e':	return 14;
		case 'f':	return 15;
		default:	return 0;
	}
}

static unsigned char hex_to_byte( char hex[2] )
{
	return (hex_to_nybble(hex[0])<<4) | hex_to_nybble(hex[1]); 
	
}

int irods_reli_md5( const char *server, const char *path, char *digest )
{
	dataObjInp_t request;
	int result;
	char *str = 0;

	rcComm_t *conn = irods_reli_connect_cache(server);
	if(!conn) return -1;

	memset(&request,0,sizeof(request));
	strcpy(request.objPath,path);
	
	debug(D_IRODS,"rcDataObjChksum %s %s",server,path);
	result = rcDataObjChksum(conn,&request,&str);
	debug(D_IRODS,"= %d",result);

	if(result<0) {
		errno = getUnixErrno(result);
		return -1;
	} else {
		/* rcDataObjChecksum returns the checksum as ASCII */
		/* but Parrot is expecting 16 bytes of binary data */
		int i;

		for(i=0;i<16;i++) {
			digest[i] = hex_to_byte(&str[2*i]);
		}

		debug(D_IRODS,"%s",str);
		free(str);
		return 0;
	}
}

#endif


